/*
Module : cpdh.cpp
Purpose: Defines the implementation for a wrapper class for the PDH API
Created: PJN / 06-02-1999
History: PJN / 13-09-1999 1. Fixed tab problem in the source code
                          2. Fixed a problem which was causing an unhandled 
                          exception being thrown when the PDH.DLL is not present on client machines..
         PJN / 25-01-2000 1. Minor update to improve the trace statements when exceptions
                          are thrown
         PJN / 24-03-2000 1. Fixed a problem with size of arrays in CPdh::EnumMachines,
                          CPdh::EnumObjects, CPdh::EnumObjectItems, CPdh::GetDefaultCounter,
                          CPdh::GetDefaultObject and CPdh::ParseInstanceName when compiled for UNICODE.
         PJN / 14-04-2001 1. Fixed a compile problem when statically linking to MFC
                          2. Updated copyright information   
                          3. Fixed a bug when calling PdhMakeCounterPath which was causing corruption
                          4. Fixed a memory leak in CPdhCounter::CPdhCounter
         PJN / 26-10-2001 1. Fixed a bug in EnumObjectItems, EnumObjects, EnumMachines and ExpandPath where 
                          unitialized data was being parsed. Thanks to Petr Stehlik for spotting the problem.
                          2. Fixed a number of functions where on Windows XP you get returned PDH_MORE_DATA even 
                          when just asking for the buffer size to allocate.
         PJN / 15-02-2003 1. Added some diagnostic code to CPDHQuery::Close() to ASSERT if any counters are
                          still present in the query when a query object is being destroyed. Some users of CPdh 
                          have reported problems with leaks in some versions of PDH.DLL when on Windows NT 4/SP6
                          when do call Add() to add a counter to a query, you must call Remove() before deleting 
                          either the Query or the counter. Failure to do so give quite a hefty memory leak. The 
                          Microsoft docs imply you don't need to do this. Thanks to Pete Williams for reporting 
                          this. Please note that this is not a bug in the CPDH classes but in PDH.dll itself (or
                          at least some versions of it).
                          2. Addition of operators in the query and counter class to return the underlying 
                          HCOUNTER / HQUERY.
         PJN / 27-12-2006 1. Updated copyright details.
                          2. Updated the documentation to use the same style as the web site.
                          3. Code now uses new C++ style casts rather than old style C casts where necessary.
                          4. Optimized CPdhException constructor code
                          5. Removed CPdhException destructor code as it was not really required
                          6. Optimized CPdhCounter constructor code
                          7. Removed CPdhCounter destructor code as it was not really required
                          8. Optimized CPdhQuery constructor code
                          9. Fix a minor copy and paste bug in CPdh::EnumObjects
                          10. Updated code to clean compile on VC 2005
                          11. CPdhQuery::Add now takes an additional BOOL parameter to specify if PdhAddCounter
                          or PdhAddEnglishCounter should be used
                          12. CPdh::GetDefaultObject now allows the data source to be specified
                          13. CPdh::EnumObjects now allows the data source to be specified
                          14. CPdh::EnumObjectItems now allows the data source to be specified
                          15. CPdhQuery::Open now allows the data source to be specified
                          16. Updated the documentation to provide a more concise description of the various
                          17. classes and their functionality.
                          18. Added support for the following PDH functions:
                          
                          PdhAddEnglishCounter,
                          PdhBindInputDataSource,
                          PdhBrowseCountersH,
                          PdhCloseLog,
                          PdhCollectQueryDataEx,
                          PdhCollectQueryDataWithTime,
                          PdhEnumLogSetNames,
                          PdhEnumMachinesH,
                          PdhEnumObjectItemsH,
                          PdhEnumObjectsH
                          PdhExpandCounterPathH,
                          PdhExpandWildCardPath
                          PdhExpandWildCardPathH,
                          PdhFormatFromRawValue,
                          PdhGetCounterTimeBase,
                          PdhGetDataSourceTimeRange,
                          PdhGetDataSourceTimeRangeH,
                          PdhGetDefaultPerfCounterH,
                          PdhGetDefaultPerfObjectH,
                          PdhGetFormattedCounterArray,
                          PdhGetLogFileSize,
                          PdhIsRealTimeQuery,
                          PdhLookupPerfIndexByName,
                          PdhLookupPerfNameByIndex,
                          PdhOpenLog
                          PdhOpenQueryH,
                          PdhReadRawLogRecord,
                          PdhSelectDataSource,
                          PdhSetDefaultRealTimeDataSource,
                          PdhSetQueryTimeRange,
                          PdhUpdateLog
                          PdhValidatePathEx
         PJN / 27-12-2007 1. CPdhException::GetErrorMessage now uses the FORMAT_MESSAGE_IGNORE_INSERTS flag. 
                          For more information please see Raymond Chen's blog at 
                          http://blogs.msdn.com/oldnewthing/archive/2007/11/28/6564257.aspx. Thanks to Alexey 
                          Kuznetsov for reporting this issue.
                          2. Minor update to display HRESULT's correctly in CPdh::ThrowPdhException
                          3. CPDHException::GetErrorMessage now uses Checked::tcsncpy_s if compiled using
                          VC 2005 or later.
         PJN / 12-01-2008 1. Updated copyright details.
                          2. Updated the code and sample app to clean compile on VC 2008  
                          3. CPdhException::GetErrorMessage now only calls the FormatMessage function once. 
                          The second call is not required following a review of the FormatMessage documentation.
                          4. CPdhException::GetErrorMessage now uses the correct return value from 
                          FormatMessage.
         PJN / 10-04-2010 1. Updated copyright details
                          2. Updated the project settings to more modern default values.
                          3. Reworked all the internal heap allocation code to use ATL::CHeapPtr
                          3. Refactored all the multi-sz parsing code into one method of CPdh
                          4. Reviewed some reported issues with CPdh::ParseInstanceName. The sample app now uses
                          a string which should parse. Thanks to David Chaplin for reporting this issue.
                          5. Some of the methods in CPdhCounter and CPdhDataSource can allocate memory. These 
                          details are now encapsulated in AllocateMemory and DeallocateMemory methods, rather than 
                          exposing raw CRT pointers in the client API.
                          6. Verified there has been no further additions to PDH from the Windows 7 SDK  
                          7. The code now only supports VC 2005 or later.          
         PJN / 08-10-2012 1. Updated copyright details
                          2. Updated the code to compile cleanly on VC 2012
                          3. Code no longer uses LoadLibrary without an absolute path when loading PDH DLL. This
                          avoids DLL planting security issues.
         PJN / 04-05-2015 1. Updated the code to clean compile on VC 2013
                          2. Updated copyright details
                          3. Reworked the classes to optionally compile without MFC. By default the classes now use STL 
                          classes and idioms but if you define CPDH_MFC_EXTENSIONS the classes will revert back to the 
                          MFC behaviour.
                          4. Added SAL annotations to all the code
                          5. Reworked CPdhDataSource::Update to take a LPCTSTR parameter
                          6. Verified there has been no further additions to PDH from the Windows 8.1 SDK
                          7. Added a new preprocessor macro value of "CPDH_NO_DYNLINK" to control inclusion of
                          "cpdhimpl.h" in cpdh.cpp.
         PJN / 22-12-2015 1. Verified there has been no further additions to PDH from the Windows 10 SDK
                          2. Updated the code to clean compile on VC 2015
         PJN / 25-03-2017 1. Updated copyright details.
                          2. Updated the code to clean compile on VC 2017
         PJN / 09-12-2017 1. Verified there has been no further additions to PDH from the latest Windows
                          10 SDK
                          2. Updated the code to compile cleanly when _ATL_NO_AUTOMATIC_NAMESPACE is
                          defined.
                          3. Replaced CString::operator LPC*STR() calls throughout the codebase with
                          CString::GetString calls
         PJN / 15-08-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean that 
                          the code will now only compile on VC 2017 or later.
                          3. Removed the code which supported the CPDH_MFC_EXTENSIONS preprocessor define
                          throughout the codebase
                          4. Replaced NULL throughout the codebase with nullptr
                          5. Replaced ATL::CHeapPtr throughout the code with std::vector
                          6. Remove usage of cpdhimpl.cpp module from the codebase. Now all calls to PDH
                          API are direct.
                          7. Added support for PdhGetDefaultPerfCounterH API.
                          8. Added support for PdhUpdateLogFileCatalog API.
         PJN / 01-09-2018 1. Fixed a number of compiler warnings when using VS 2017 15.8.1
         PJN / 21-05-2019 1. Updated copyright details.
                          2. Updated the code to clean compile on VC 2019
         PJN / 29-12-2019 1. Fixed various Clang-Tidy static code analysis warnings in the code.
         PJN / 21-03-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 18-03-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations.

Copyright (c) 1999 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


///////////////////////////////// Includes ////////////////////////////////////

#include "stdafx.h"
#include "cpdh.h"
#include <pdhmsg.h>


//////////////////////////////// Macros / Defines /////////////////////////////


#if (_WIN32_WINDOWS < 0x0600 || _WIN32_WINNT < 0x0600 || (defined(NTDDI_VERSION) && NTDDI_VERSION < NTDDI_VISTA))
#pragma message("CPdh as of v1.18 requires NTDDI_VERSION to be >= NTDDI_VISTA")
#endif //#if (NTDDI_VERSION < NTDDI_VISTA)

#pragma comment(lib, "pdh.lib")


//////////////////////////////// Implementation ///////////////////////////////

CPdhException::CPdhException(_In_ PDH_STATUS Status) noexcept : m_Status{Status}
{
}

#pragma warning(suppress: 26429)
BOOL CPdhException::GetErrorMessage(_Out_writes_z_(nMaxError) LPTSTR lpszError, _In_ UINT nMaxError, _Out_opt_ PUINT pnHelpContext)
{
  //Validate out parameters
#pragma warning(suppress: 26477)
  ATLASSERT(lpszError != nullptr);

  //We don't support help context at the moment
  if (pnHelpContext != nullptr)
    *pnHelpContext = 0;

  //What will be the return value from this function (assume the worst)
  BOOL bSuccess{FALSE};

  //First try getting the error text from the PDH module
  LPTSTR lpBuffer{nullptr};
#pragma warning(suppress: 26490)
  const DWORD dwReturn{FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_HMODULE | FORMAT_MESSAGE_IGNORE_INSERTS,
                                     GetModuleHandle(_T("PDH.DLL")), m_Status, MAKELANGID(LANG_NEUTRAL, SUBLANG_SYS_DEFAULT), reinterpret_cast<LPTSTR>(&lpBuffer), 0, nullptr)};
  if (dwReturn == 0)
    *lpszError = _T('\0');
  else
  {
    bSuccess = TRUE;
    ATL::Checked::tcsncpy_s(lpszError, nMaxError, lpBuffer, _TRUNCATE);
    LocalFree(lpBuffer);
  }

  return bSuccess;
}


CPdhCounter::CPdhCounter(_In_ LPCTSTR pszFullCounterPath, _In_ DWORD_PTR dwUserData) : m_hCounter{nullptr},
                                                                                       m_sCounterPath{pszFullCounterPath},
                                                                                       m_dwUserData{dwUserData}
{
}

CPdhCounter::CPdhCounter(_In_ PDH_COUNTER_PATH_ELEMENTS& CounterPathElements, _In_ DWORD_PTR dwUserData) : m_hCounter{nullptr},
                                                                                                           m_dwUserData{0}
{
  //First determine the size of the string to allocate
  DWORD dwBufferSize{0};
  PDH_STATUS Status{PdhMakeCounterPath(&CounterPathElements, nullptr, &dwBufferSize, 0)};
  //Throw an exception if anything goes wrong
#pragma warning(suppress: 26472)
  if (Status != static_cast<PDH_STATUS>(PDH_MORE_DATA))
    CPdh::ThrowPdhException(Status);

  //Call again with a newly allocated buffer
  std::vector<TCHAR> pszData{dwBufferSize, std::allocator<TCHAR>{}};
  Status = PdhMakeCounterPath(&CounterPathElements, pszData.data(), &dwBufferSize, 0);
  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    CPdh::ThrowPdhException(Status);

  //Initialize our member variables
  m_hCounter = nullptr;
  m_dwUserData = dwUserData;
  m_sCounterPath = pszData.data();
}

DWORD CPdhCounter::GetRawValue(_Out_ PDH_RAW_COUNTER& value)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hCounter != nullptr);

  DWORD dwType{0};
  const PDH_STATUS Status{PdhGetRawCounterValue(m_hCounter, &dwType, &value)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    CPdh::ThrowPdhException(Status);

  return dwType;
}

void CPdhCounter::SplitPath(_Inout_ CPDHString& sMachineName, _Inout_ CPDHString& sObjectName, _Inout_ CPDHString& sInstanceName,
                            _Inout_ CPDHString& sParentInstance, _Inout_ DWORD& dwInstanceIndex, _Inout_ CPDHString& sCounterName)
{
  //First call the function to determine the size of buffer required
  DWORD dwBufferSize{0};
  PDH_STATUS Status{PdhParseCounterPath(m_sCounterPath.c_str(), nullptr, &dwBufferSize, 0)};
#pragma warning(suppress: 26472)
  if (Status == static_cast<PDH_STATUS>(PDH_MORE_DATA))
  {
    //Call again with a newly allocated buffer
    std::vector<BYTE> elements{dwBufferSize, std::allocator<BYTE>{}};
#pragma warning(suppress: 26490)
    auto pszBuffer{reinterpret_cast<PDH_COUNTER_PATH_ELEMENTS*>(elements.data())};
#pragma warning(suppress: 26486)
    Status = PdhParseCounterPath(m_sCounterPath.c_str(), pszBuffer, &dwBufferSize, 0);
    if (Status == ERROR_SUCCESS)
    {
      //Assign the data into the strings sent into us
      if (pszBuffer->szMachineName != nullptr)
#pragma warning(suppress: 26486)
        sMachineName = pszBuffer->szMachineName;
      else
        sMachineName.clear();
      if (pszBuffer->szObjectName != nullptr)
#pragma warning(suppress: 26486)
        sObjectName = pszBuffer->szObjectName;
      else
        sObjectName.clear();
      if (pszBuffer->szInstanceName != nullptr)
#pragma warning(suppress: 26486)
        sInstanceName = pszBuffer->szInstanceName;
      else
        sInstanceName.clear();
      if (pszBuffer->szParentInstance != nullptr)
#pragma warning(suppress: 26486)
        sParentInstance = pszBuffer->szParentInstance;
      else
        sParentInstance.clear();
      dwInstanceIndex = pszBuffer->dwInstanceIndex;
      if (pszBuffer->szCounterName != nullptr)
#pragma warning(suppress: 26486)
        sCounterName = pszBuffer->szCounterName;
      else
        sCounterName.clear();
    }
  }

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    CPdh::ThrowPdhException(Status);
}

void CPdhCounter::GetInfo(_Inout_ CPdhCounterInfo& info)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hCounter != nullptr);

  //First call the function to determine the size of buffer required
  DWORD dwBufferSize{0};
  PDH_STATUS Status{PdhGetCounterInfo(m_hCounter, TRUE, &dwBufferSize, nullptr)};
#pragma warning(suppress: 26472)
  if (Status == static_cast<PDH_STATUS>(PDH_MORE_DATA))
  {
    //Call again with the newly allocated buffer
    std::vector<BYTE> szBuffer{dwBufferSize, std::allocator<BYTE>{}};
#pragma warning(suppress: 26490)
    auto pszBuffer{reinterpret_cast<PPDH_COUNTER_INFO>(szBuffer.data())};
#pragma warning(suppress: 26486)
    Status = PdhGetCounterInfo(m_hCounter, TRUE, &dwBufferSize, pszBuffer);
    if (Status == ERROR_SUCCESS)
    {
      //Assign the data into the class sent into us
      info.m_dwType = pszBuffer->dwType;
      info.m_CVersion = pszBuffer->CVersion;
      info.m_CStatus = pszBuffer->CStatus;
      info.m_lScale = pszBuffer->lScale;
      info.m_lDefaultScale = pszBuffer->lDefaultScale;
      info.m_dwUserData = pszBuffer->dwUserData;
      info.m_dwQueryUserData = pszBuffer->dwQueryUserData;
      if (pszBuffer->szFullPath != nullptr)
#pragma warning(suppress: 26486)
        info.m_sFullPath = pszBuffer->szFullPath;
      else
        info.m_sFullPath.clear();
      if (pszBuffer->szMachineName != nullptr)
#pragma warning(suppress: 26486)
        info.m_sMachineName = pszBuffer->szMachineName;
      else
        info.m_sMachineName.clear();
      if (pszBuffer->szObjectName != nullptr)
#pragma warning(suppress: 26486)
        info.m_sInstanceName = pszBuffer->szObjectName;
      else
        info.m_sInstanceName.clear();
      if (pszBuffer->szParentInstance != nullptr)
#pragma warning(suppress: 26486)
        info.m_sParentInstance = pszBuffer->szParentInstance;
      else
        info.m_sParentInstance.clear();
      info.m_dwInstanceIndex = pszBuffer->dwInstanceIndex;
      if (pszBuffer->szCounterName != nullptr)
#pragma warning(suppress: 26486)
        info.m_sCounterName = pszBuffer->szCounterName;
      else
        info.m_sCounterName.clear();
      if (pszBuffer->szExplainText != nullptr)
#pragma warning(suppress: 26486)
        info.m_sExplainText = pszBuffer->szExplainText;
      else
        info.m_sExplainText.clear();
    }
  }

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    CPdh::ThrowPdhException(Status);
}

void CPdhCounter::CalculateFromRawValue(_In_ DWORD dwFormat, _In_ PDH_RAW_COUNTER& rawValue1,
                                        _In_ PDH_RAW_COUNTER& rawValue2, _Out_ PDH_FMT_COUNTERVALUE& fmtValue)
{
  //Validtae our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hCounter != nullptr);

  const PDH_STATUS Status{PdhCalculateCounterFromRawValue(m_hCounter, dwFormat, &rawValue1, &rawValue2, &fmtValue)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    CPdh::ThrowPdhException(Status);
}

DWORD CPdhCounter::GetFormattedValue(_In_ DWORD dwFormat, _Out_ PDH_FMT_COUNTERVALUE& value)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hCounter != nullptr);

  DWORD dwType{0};
  const PDH_STATUS Status{PdhGetFormattedCounterValue(m_hCounter, dwFormat, &dwType, &value)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    CPdh::ThrowPdhException(Status);

  return dwType;
}

void CPdhCounter::ComputeStatistics(_In_ DWORD dwFormat, _In_ DWORD dwFirstEntry, _In_ DWORD dwNumEntries,
                                    _In_ PPDH_RAW_COUNTER lpRawValueArray, _Out_ PPDH_STATISTICS data)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hCounter != nullptr);

  const PDH_STATUS Status{PdhComputeCounterStatistics(m_hCounter, dwFormat, dwFirstEntry, dwNumEntries, lpRawValueArray, data)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    CPdh::ThrowPdhException(Status);
}

void CPdhCounter::SetScaleFactor(_In_ LONG lFactor)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hCounter != nullptr);

  const PDH_STATUS Status{PdhSetCounterScaleFactor(m_hCounter, lFactor)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    CPdh::ThrowPdhException(Status);
}

LONGLONG CPdhCounter::GetTimeBase()
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hCounter != nullptr);

  LONGLONG nTimeBase{0};
  const PDH_STATUS Status{PdhGetCounterTimeBase(m_hCounter, &nTimeBase)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    CPdh::ThrowPdhException(Status);

  return nTimeBase;
}

PPDH_FMT_COUNTERVALUE_ITEM CPdhCounter::GetFormattedCounterArray(_In_ DWORD dwFormat, _Out_ DWORD& dwItemsInArray)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hCounter != nullptr);

  //What will be the return value from this function
  PPDH_FMT_COUNTERVALUE_ITEM pItems{nullptr};
  dwItemsInArray = 0;

  //First call the function to determine the size of buffer required
  DWORD dwBufferSize{0};
  PDH_STATUS Status{PdhGetFormattedCounterArray(m_hCounter, dwFormat, &dwBufferSize, &dwItemsInArray, nullptr)};
#pragma warning(suppress: 26472)
  if (Status == static_cast<PDH_STATUS>(PDH_MORE_DATA))
  {
    //Call again with the newly allocated buffer (client code is expected to do a DeallocateMemory on the returned value)
    pItems = static_cast<PPDH_FMT_COUNTERVALUE_ITEM>(AllocateMemory(dwBufferSize));
    if (pItems == nullptr)
      CPdh::ThrowPdhException(PDH_MEMORY_ALLOCATION_FAILURE);
#pragma warning(suppress: 26486)
    Status = PdhGetFormattedCounterArray(m_hCounter, dwFormat, &dwBufferSize, &dwItemsInArray, pItems);
  }

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
  {
    //Free up the memory before we throw the exception
    if (pItems != nullptr)
      DeallocateMemory(pItems);
    CPdh::ThrowPdhException(Status);
  }

#pragma warning(suppress: 6001)
  return pItems;
}

PPDH_RAW_COUNTER_ITEM CPdhCounter::GetRawCounterArray(_Out_ DWORD& dwItemsInArray)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hCounter != nullptr);

  //Initialize the output parameter value
  dwItemsInArray = 0;

  //What will be the return value from this function
  PPDH_RAW_COUNTER_ITEM pItems{nullptr};

  //First call the function to determine the size of buffer required
  DWORD dwBufferSize{0};
  PDH_STATUS Status{PdhGetRawCounterArray(m_hCounter, &dwBufferSize, &dwItemsInArray, nullptr)};
#pragma warning(suppress: 26472)
  if (Status == static_cast<PDH_STATUS>(PDH_MORE_DATA))
  {
    //Call again with the newly allocated buffer (client code is expected to do a DeallocateMemory on the returned value)
    pItems = static_cast<PPDH_RAW_COUNTER_ITEM>(AllocateMemory(dwBufferSize));
    if (pItems == nullptr)
      CPdh::ThrowPdhException(PDH_MEMORY_ALLOCATION_FAILURE);
#pragma warning(suppress: 26486)
    Status = PdhGetRawCounterArray(m_hCounter, &dwBufferSize, &dwItemsInArray, pItems);
  }

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
  {
    //Free up the memory before we throw the exception
    if (pItems != nullptr)
      DeallocateMemory(pItems);
    CPdh::ThrowPdhException(Status);
  }

#pragma warning(suppress: 6001)
  return pItems;
}

LPVOID CPdhCounter::AllocateMemory(_In_ DWORD dwBytes) noexcept
{
  //By default we allocate from the Process Heap
  return HeapAlloc(GetProcessHeap(), 0, dwBytes);
}

void CPdhCounter::DeallocateMemory(__drv_freesMem(Mem) _Frees_ptr_opt_ void* pData) noexcept
{
  //By default we deallocate from the Process Heap
  HeapFree(GetProcessHeap(), 0, pData);
}


CPdhQuery::CPdhQuery() noexcept : m_hQuery{nullptr},
                                  m_dwCountersInThisQuery{0}
{
}

CPdhQuery::~CPdhQuery()
{
  //Ensure we don't leak a query handle
  Close();
}

void CPdhQuery::Open(_In_opt_ LPCTSTR szDataSource, _In_ DWORD_PTR dwUserData)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hQuery == nullptr);

  const PDH_STATUS Status{PdhOpenQuery(szDataSource, dwUserData, &m_hQuery)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    CPdh::ThrowPdhException(Status);
}

void CPdhQuery::OpenH(_In_opt_ PDH_HLOG hDataSource, _In_ DWORD_PTR dwUserData)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hQuery == nullptr);

  const PDH_STATUS Status{PdhOpenQueryH(hDataSource, dwUserData, &m_hQuery)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    CPdh::ThrowPdhException(Status);
}

BOOL CPdhQuery::IsRealTime() noexcept
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hQuery != nullptr); //Did you forget to call Open

  return PdhIsRealTimeQuery(m_hQuery);
}

void CPdhQuery::Close() noexcept
{
  if (m_hQuery != nullptr)
  {
  #ifdef _DEBUG
    //To avoid memory leaks in some versions of PDH.DLL, you should ensure that all counters
    //are removed from a query by calling CPdhQuery::Remove before you close the query via
    //this function CPdhQuery::Close or when it is automatically called via the CPdhQuery
    //destructor
#pragma warning(suppress: 26477)
    ATLASSERT(m_dwCountersInThisQuery == 0);
  #endif //#ifdef _DEBUG

    PdhCloseQuery(m_hQuery);
    m_hQuery = nullptr;
  }
}

void CPdhQuery::Add(_Inout_ CPdhCounter& counter, _In_ BOOL bUseEnglishName)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hQuery != nullptr); //Did you forget to call Open

  const PDH_STATUS Status{bUseEnglishName ? PdhAddEnglishCounter(m_hQuery, counter.m_sCounterPath.c_str(), counter.m_dwUserData, &counter.m_hCounter) :
                                            PdhAddCounter(m_hQuery, counter.m_sCounterPath.c_str(), counter.m_dwUserData, &counter.m_hCounter)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    CPdh::ThrowPdhException(Status);
  else
    ++m_dwCountersInThisQuery;
}

void CPdhQuery::Remove(_Inout_ CPdhCounter& counter)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hQuery != nullptr); //Did you forget to call Open

  const PDH_STATUS Status{PdhRemoveCounter(counter.m_hCounter)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    CPdh::ThrowPdhException(Status);
  else
    --m_dwCountersInThisQuery;
}

void CPdhQuery::Collect()
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hQuery != nullptr); //Did you forget to call Open

  const PDH_STATUS Status{PdhCollectQueryData(m_hQuery)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    CPdh::ThrowPdhException(Status);
}
 
void CPdhQuery::Collect(_In_ DWORD dwIntervalTime, _In_ HANDLE hNewDataEvent)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hQuery != nullptr); //Did you forget to call Open

  const PDH_STATUS Status{PdhCollectQueryDataEx(m_hQuery, dwIntervalTime, hNewDataEvent)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    CPdh::ThrowPdhException(Status);
}

LONGLONG CPdhQuery::CollectWithTime()
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hQuery != nullptr); //Did you forget to call Open

  LONGLONG llTimeStamp{0};
  const PDH_STATUS Status{PdhCollectQueryDataWithTime(m_hQuery, &llTimeStamp)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    CPdh::ThrowPdhException(Status);

  return llTimeStamp;
}

void CPdhQuery::SetTimeRange(_In_ PPDH_TIME_INFO pInfo)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hQuery != nullptr); //Did you forget to call Open

  const PDH_STATUS Status{PdhSetQueryTimeRange(m_hQuery, pInfo)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    CPdh::ThrowPdhException(Status);
}


void CPdh::ValidatePath(_In_ LPCTSTR pszFullCounterPath)
{
  const PDH_STATUS Status{PdhValidatePath(pszFullCounterPath)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);
}

void CPdh::ValidatePath(_In_opt_ PDH_HLOG hDataSource, _In_ LPCTSTR szFullPathBuffer)
{
  const PDH_STATUS Status{PdhValidatePathEx(hDataSource, szFullPathBuffer)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);
}

CPDHStringArray CPdh::ParseMultiSZString(_In_ _NullNull_terminated_ LPCTSTR pszValue)
{
  //Reset the output parameter array
  CPDHStringArray values;

  if (pszValue == nullptr)
    return values;

#pragma warning(suppress: 26429)
  LPCTSTR pszTemp{pszValue};
#pragma warning(suppress: 26489)
  while (*pszTemp)
  {
#pragma warning(suppress: 26486 26489)
    values.emplace_back(pszTemp);
#pragma warning(suppress: 26481 26486)
    pszTemp += _tcslen(pszTemp);
#pragma warning(suppress: 26481)
    pszTemp++;
  }

  return values;
}

CPDHStringArray CPdh::ExpandPath(_In_ LPCTSTR pszWildCardPath)
{
  //What will be the return value from this method
  CPDHStringArray expandedPaths;

  //First call the function to determine the size of buffer required
  DWORD dwBufferSize{0};
  PDH_STATUS Status{PdhExpandCounterPath(pszWildCardPath, nullptr, &dwBufferSize)};
#pragma warning(suppress: 26472)
  if (Status == static_cast<PDH_STATUS>(PDH_MORE_DATA))
  {
    //Call again with the newly allocated buffer
    dwBufferSize++; //Workaround to suit PdhExpandCounterPath behaviour on XP
    std::vector<TCHAR> mszExpandedPathList{dwBufferSize, std::allocator<TCHAR>{}};
    Status = PdhExpandCounterPath(pszWildCardPath, mszExpandedPathList.data(), &dwBufferSize);

    //Parse out the MULTI_SZ strings and add each item to the string array
    if ((Status == ERROR_SUCCESS) && dwBufferSize)
      expandedPaths = ParseMultiSZString(mszExpandedPathList.data());
  }

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);

  return expandedPaths;
}

void CPdh::ConnectMachine(_In_opt_ LPCTSTR pszMachineName)
{
  const PDH_STATUS Status{PdhConnectMachine(pszMachineName)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);
}

CPDHStringArray CPdh::EnumMachines(_In_opt_ LPCTSTR pszDataSource)
{
  //What will be the return value from this method
  CPDHStringArray machineNames;

  //First call the function to determine the size of buffer required
  DWORD dwBufferSize{0};
  PDH_STATUS Status{PdhEnumMachines(pszDataSource, nullptr, &dwBufferSize)};
#pragma warning(suppress: 26472)
  if (Status == static_cast<PDH_STATUS>(PDH_MORE_DATA))
  {
    //Call again with the newly allocated buffer
    std::vector<TCHAR> mszMachineNameList{dwBufferSize, std::allocator<TCHAR>{}};
    Status = PdhEnumMachines(pszDataSource, mszMachineNameList.data(), &dwBufferSize);

    //Parse out the MULTI_SZ strings and add each item to the string array
    if ((Status == ERROR_SUCCESS) && dwBufferSize)
      machineNames = ParseMultiSZString(mszMachineNameList.data());
  }

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);

  return machineNames;
}

CPDHStringArray CPdh::EnumMachinesH(_In_opt_ PDH_HLOG hDataSource)
{
  //What will be the return value from this method
  CPDHStringArray machineNames;

  //First call the function to determine the size of buffer required
  DWORD dwBufferSize{0};
  PDH_STATUS Status{PdhEnumMachinesH(hDataSource, nullptr, &dwBufferSize)};
#pragma warning(suppress: 26472)
  if (Status == static_cast<PDH_STATUS>(PDH_MORE_DATA))
  {
    //Call again with the newly allocated buffer
    std::vector<TCHAR> mszMachineNameList{dwBufferSize, std::allocator<TCHAR>{}};
    Status = PdhEnumMachinesH(hDataSource, mszMachineNameList.data(), &dwBufferSize);

    //Parse out the MULTI_SZ strings and add each item to the string array
    if ((Status == ERROR_SUCCESS) && dwBufferSize)
      machineNames = ParseMultiSZString(mszMachineNameList.data());
  }

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);

  return machineNames;
}

CPDHStringArray CPdh::EnumObjects(_In_opt_ LPCTSTR pszDataSource, _In_opt_ LPCTSTR pszMachineName, _In_ DWORD dwDetailLevel, _In_ BOOL bRefresh)
{
  //What will be the return value from this method
  CPDHStringArray objects;

  //First call the function to determine the size of buffer required
  DWORD dwBufferSize{0};
  PDH_STATUS Status{PdhEnumObjects(pszDataSource, pszMachineName, nullptr, &dwBufferSize, dwDetailLevel, bRefresh)};
#pragma warning(suppress: 26472)
  if (Status == static_cast<PDH_STATUS>(PDH_MORE_DATA))
  {
    //Call again with the newly allocated buffer
    std::vector<TCHAR> mszObjectList{dwBufferSize, std::allocator<TCHAR>{}};
    Status = PdhEnumObjects(pszDataSource, pszMachineName, mszObjectList.data(), &dwBufferSize, dwDetailLevel, bRefresh);

    //Parse out the MULTI_SZ strings and add each item to the string array
    if ((Status == ERROR_SUCCESS) && dwBufferSize)
      objects = ParseMultiSZString(mszObjectList.data());
  }

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);

  return objects;
}

CPDHStringArray CPdh::EnumObjectsH(_In_opt_ PDH_HLOG hDataSource, _In_opt_ LPCTSTR pszMachineName, _In_ DWORD dwDetailLevel, _In_ BOOL bRefresh)
{
  //What will be the return value from this method
  CPDHStringArray objects;

  //First call the function to determine the size of buffer required
  DWORD dwBufferSize{0};
  PDH_STATUS Status{PdhEnumObjectsH(hDataSource, pszMachineName, nullptr, &dwBufferSize, dwDetailLevel, bRefresh)};
#pragma warning(suppress: 26472)
  if (Status == static_cast<PDH_STATUS>(PDH_MORE_DATA))
  {
    //Call again with the newly allocated buffer
    std::vector<TCHAR> mszObjectList{dwBufferSize, std::allocator<TCHAR>{}};
    Status = PdhEnumObjectsH(hDataSource, pszMachineName, mszObjectList.data(), &dwBufferSize, dwDetailLevel, bRefresh);

    //Parse out the MULTI_SZ strings and add each item to the string array
    if ((Status == ERROR_SUCCESS) && dwBufferSize)
      objects = ParseMultiSZString(mszObjectList.data());
  }

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);

  return objects;
}

CPDHStringArray CPdh::EnumLogSetNames(_In_ LPCTSTR pszDataSource)
{
  //What will be the return value from this method
  CPDHStringArray LogSetNames;

  //First call the function to determine the size of buffer required
  DWORD dwBufferSize{0};
  PDH_STATUS Status{PdhEnumLogSetNames(pszDataSource, nullptr, &dwBufferSize)};
#pragma warning(suppress: 26472)
  if (Status == static_cast<PDH_STATUS>(PDH_MORE_DATA))
  {
    //Call again with the newly allocated buffer
    std::vector<TCHAR> mszObjectList{dwBufferSize, std::allocator<TCHAR>{}};
    Status = PdhEnumLogSetNames(pszDataSource, mszObjectList.data(), &dwBufferSize);

    //Parse out the MULTI_SZ strings and add each item to the string array
    if ((Status == ERROR_SUCCESS) && dwBufferSize)
      LogSetNames = ParseMultiSZString(mszObjectList.data());
  }

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);

  return LogSetNames;
}

void CPdh::EnumObjectItems(_In_opt_ LPCTSTR szDataSource, _In_opt_ LPCTSTR szMachineName, _In_ LPCTSTR szObjectName, _Out_ CPDHStringArray& counters, _Out_ CPDHStringArray& instances, _In_ DWORD dwDetailLevel, _In_ DWORD dwFlags)
{
  //Default the output parameters
  counters.clear();
  instances.clear();

  //First call the function to determine the size of buffer required
  DWORD dwCounterBufferSize{0};
  DWORD dwInstanceBufferSize{0};
  PDH_STATUS Status{PdhEnumObjectItems(szDataSource, szMachineName, szObjectName, nullptr, &dwCounterBufferSize, nullptr, &dwInstanceBufferSize, dwDetailLevel, dwFlags)};
#pragma warning(suppress: 26472)
  if (Status == static_cast<PDH_STATUS>(PDH_MORE_DATA))
  {
    //Call again with the newly allocated buffer
    std::vector<TCHAR> pszCounterBuffer{dwCounterBufferSize, std::allocator<TCHAR>{}};
    std::vector<TCHAR> pszInstanceBuffer{dwInstanceBufferSize, std::allocator<TCHAR>{}};
    Status = PdhEnumObjectItems(szDataSource, szMachineName, szObjectName, pszCounterBuffer.data(), &dwCounterBufferSize, pszInstanceBuffer.data(), &dwInstanceBufferSize, dwDetailLevel, dwFlags);
    if (Status == ERROR_SUCCESS)
    {
      //Parse out the MULTI_SZ strings which form the Counter string array
      if (dwCounterBufferSize)
        counters = ParseMultiSZString(pszCounterBuffer.data());

      //Parse out the MULTI_SZ strings which form the Instances string array
      if (dwInstanceBufferSize)
        instances = ParseMultiSZString(pszInstanceBuffer.data());
    }
  }

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);
}

void CPdh::EnumObjectItemsH(_In_opt_ PDH_HLOG hDataSource, _In_opt_ LPCTSTR szMachineName, _In_ LPCTSTR szObjectName, _Out_ CPDHStringArray& counters, _Out_ CPDHStringArray& instances, _In_ DWORD dwDetailLevel, _In_ DWORD dwFlags)
{
  //Default the output parameters
  counters.clear();
  instances.clear();

  //First call the function to determine the size of buffer required
  DWORD dwCounterBufferSize{0};
  DWORD dwInstanceBufferSize{0};
  PDH_STATUS Status{PdhEnumObjectItemsH(hDataSource, szMachineName, szObjectName, nullptr, &dwCounterBufferSize, nullptr, &dwInstanceBufferSize, dwDetailLevel, dwFlags)};
#pragma warning(suppress: 26472)
  if (Status == static_cast<PDH_STATUS>(PDH_MORE_DATA))
  {
    //Call again with the newly allocated buffer
    std::vector<TCHAR> pszCounterBuffer{dwCounterBufferSize, std::allocator<TCHAR>{}};
    std::vector<TCHAR> pszInstanceBuffer{dwInstanceBufferSize, std::allocator<TCHAR>{}};
    Status = PdhEnumObjectItemsH(hDataSource, szMachineName, szObjectName, pszCounterBuffer.data(), &dwCounterBufferSize, pszInstanceBuffer.data(), &dwInstanceBufferSize, dwDetailLevel, dwFlags);
    if (Status == ERROR_SUCCESS)
    {
      //Parse out the MULTI_SZ strings which form the Counter string array
      if (dwCounterBufferSize)
        counters = ParseMultiSZString(pszCounterBuffer.data());

      //Parse out the MULTI_SZ strings which form the Instances string array
      if (dwInstanceBufferSize)
        instances = ParseMultiSZString(pszInstanceBuffer.data());
    }
  }

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);
}

void CPdh::GetDefaultCounter(_In_opt_ LPCTSTR szMachineName, _In_ LPCTSTR szObjectName, _Inout_ CPDHString& sDefaultCounterName)
{
  //First call the function to determine the size of buffer required
  DWORD dwBufferSize{0};
  PDH_STATUS Status{PdhGetDefaultPerfCounter(nullptr, szMachineName, szObjectName, nullptr, &dwBufferSize)};
#pragma warning(suppress: 26472)
  if (Status == static_cast<PDH_STATUS>(PDH_MORE_DATA))
  {
    //Call again with the newly allocated buffer
    std::vector<TCHAR> pszBuffer{dwBufferSize, std::allocator<TCHAR>{}};
    Status = PdhGetDefaultPerfCounter(nullptr, szMachineName, szObjectName, pszBuffer.data(), &dwBufferSize);
    if (Status == ERROR_SUCCESS)
    {
      //Assign the data into the string sent into us
      sDefaultCounterName = pszBuffer.data();
    }
  }

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);
}

void CPdh::GetDefaultObject(_In_opt_ LPCTSTR szDataSource, _In_opt_ LPCTSTR szMachineName, _Inout_ CPDHString& sDefaultObjectName)
{
  //Default the output parameter
  sDefaultObjectName.clear();

  //First call the function to determine the size of buffer required
  DWORD dwBufferSize{0};
  PDH_STATUS Status{PdhGetDefaultPerfObject(szDataSource, szMachineName, nullptr, &dwBufferSize)};
#pragma warning(suppress: 26472)
  if (Status == static_cast<PDH_STATUS>(PDH_MORE_DATA))
  {
    //Call again with the newly allocated buffer
    std::vector<TCHAR> pszBuffer{dwBufferSize, std::allocator<TCHAR>{}};
    Status = PdhGetDefaultPerfObject(szDataSource, szMachineName, pszBuffer.data(), &dwBufferSize);
    if (Status == ERROR_SUCCESS)
    {
      //Assign the data into the string sent into us
      sDefaultObjectName = pszBuffer.data();
    }
  }

  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);
}

void CPdh::GetDefaultObjectH(_In_opt_ PDH_HLOG hDataSource, _In_opt_ LPCTSTR szMachineName, _Inout_ CPDHString& sDefaultObjectName)
{
  //Default the output parameter
  sDefaultObjectName.clear();

  //First call the function to determine the size of buffer required
  DWORD dwBufferSize{0};
  PDH_STATUS Status{PdhGetDefaultPerfObjectH(hDataSource, szMachineName, nullptr, &dwBufferSize)};
#pragma warning(suppress: 26472)
  if (Status == static_cast<PDH_STATUS>(PDH_MORE_DATA))
  {
    //Call again with the newly allocated buffer
    std::vector<TCHAR> pszBuffer{dwBufferSize, std::allocator<TCHAR>{}};
    Status = PdhGetDefaultPerfObjectH(hDataSource, szMachineName, pszBuffer.data(), &dwBufferSize);
    if (Status == ERROR_SUCCESS)
    {
      //Assign the data into the string sent into us
      sDefaultObjectName = pszBuffer.data();
    }
  }

  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);
}

void CPdh::BrowseCounters(_In_ PDH_BROWSE_DLG_CONFIG& BrowseDlgData)
{
  const PDH_STATUS Status{PdhBrowseCounters(&BrowseDlgData)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);
}

void CPdh::BrowseCounters(_In_ PDH_BROWSE_DLG_CONFIG_H& BrowseDlgData)
{
  const PDH_STATUS Status{PdhBrowseCountersH(&BrowseDlgData)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);
}

void CPdh::ParseInstanceName(_In_ LPCTSTR szInstanceString, _Inout_ CPDHString& sInstanceName, _Inout_ CPDHString& sParentName, _Out_ DWORD& Index)
{
  //Default the output parameters
  sInstanceName.clear();
  sParentName.clear();

  //First call the function to determine the size of buffer required
  DWORD dwInstanceNameBufferSize{0};
  DWORD dwParentNameBufferSize{0};
  PDH_STATUS Status{PdhParseInstanceName(szInstanceString, nullptr, &dwInstanceNameBufferSize, nullptr, &dwParentNameBufferSize, &Index)};
#pragma warning(suppress: 26472)
  if (Status == static_cast<PDH_STATUS>(PDH_MORE_DATA))
  {
    //Call again with the newly allocated buffer
    std::vector<TCHAR> pszInstanceName{dwInstanceNameBufferSize, std::allocator<TCHAR>{}};
    std::vector<TCHAR> pszParentName{dwParentNameBufferSize, std::allocator<TCHAR>{}};
    Status = PdhParseInstanceName(szInstanceString, pszInstanceName.data(), &dwInstanceNameBufferSize, pszParentName.data(), &dwParentNameBufferSize, &Index);
    if (Status == ERROR_SUCCESS)
    {
      //Assign the data into the strings sent into us
      sInstanceName = pszInstanceName.data();
      sParentName = pszParentName.data();
    }
  }

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);
}

void CPdh::FormatFromRawValue(_In_ DWORD dwCounterType, _In_ DWORD dwFormat, _In_ LONGLONG* pTimeBase, _In_ PPDH_RAW_COUNTER pRawValue1, _In_ PPDH_RAW_COUNTER pRawValue2, _Out_ PPDH_FMT_COUNTERVALUE pFmtValue)
{
  const PDH_STATUS Status{PdhFormatFromRawValue(dwCounterType, dwFormat, pTimeBase, pRawValue1, pRawValue2, pFmtValue)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);
}

void CPdh::ThrowPdhException(_In_ PDH_STATUS Status)
{
  ATLTRACE(_T("Warning: throwing CPdhException for error 0x%08X\n"), Status);
  CPdhException e(Status);
  throw e;
}

CPDHString CPdh::SelectDataSource(_In_ HWND hWndOwner, _In_ DWORD dwFlags)
{
  DWORD dwBufferSize{4096};
  std::vector<TCHAR> sDataSource{dwBufferSize, std::allocator<TCHAR>{}};
  const PDH_STATUS Status{PdhSelectDataSource(hWndOwner, dwFlags, sDataSource.data(), &dwBufferSize)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);

  return sDataSource.data();
}

void CPdh::GetDataSourceTimeRange(_In_opt_ LPCTSTR szDataSource, _Out_ LPDWORD pdwNumEntries, _Out_writes_bytes_(*pdwBufferSize) PPDH_TIME_INFO pInfo, _Inout_ LPDWORD pdwBufferSize)
{
  const PDH_STATUS Status{PdhGetDataSourceTimeRange(szDataSource, pdwNumEntries, pInfo, pdwBufferSize)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);
}

void CPdh::GetDataSourceTimeRangeH(_Inout_opt_ PDH_HLOG hDataSource, _Out_ LPDWORD pdwNumEntries, _Out_writes_bytes_(*pdwBufferSize) PPDH_TIME_INFO pInfo, _Inout_ LPDWORD pdwBufferSize)
{
  const PDH_STATUS Status{PdhGetDataSourceTimeRangeH(hDataSource, pdwNumEntries, pInfo, pdwBufferSize)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);
}

CPDHStringArray CPdh::ExpandWildCardPath(_In_opt_ LPCTSTR szDataSource, _In_ LPCTSTR szWildCardPath, _In_ DWORD dwFlags)
{
  //What will be the return value from this method
  CPDHStringArray expandedList;

  //First call the function to determine the size of buffer required
  DWORD dwPathListBufferSize{0};
  PDH_STATUS Status{PdhExpandWildCardPath(szDataSource, szWildCardPath, nullptr, &dwPathListBufferSize, dwFlags)};
#pragma warning(suppress: 26472)
  if (Status == static_cast<PDH_STATUS>(PDH_MORE_DATA))
  {
    //Call again with the newly allocated buffer
    std::vector<TCHAR> mszExpandedPathList{dwPathListBufferSize, std::allocator<TCHAR>{}};
    Status = PdhExpandWildCardPath(szDataSource, szWildCardPath, mszExpandedPathList.data(), &dwPathListBufferSize, dwFlags);

    //Parse out the MULTI_SZ strings which form the Counter string array
    if ((Status == ERROR_SUCCESS) && dwPathListBufferSize)
      expandedList = ParseMultiSZString(mszExpandedPathList.data());
  }
  
  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);

  return expandedList;
}

CPDHStringArray CPdh::ExpandWildCardPathH(_In_opt_ PDH_HLOG hDataSource, _In_ LPCTSTR szWildCardPath, _In_ DWORD dwFlags)
{
  //What will be the return value from this method
  CPDHStringArray expandedList;

  //First call the function to determine the size of buffer required
  DWORD dwPathListBufferSize{0};
  PDH_STATUS Status{PdhExpandWildCardPathH(hDataSource, szWildCardPath, nullptr, &dwPathListBufferSize, dwFlags)};
#pragma warning(suppress: 26472)
  if (Status == static_cast<PDH_STATUS>(PDH_MORE_DATA))
  {
    //Call again with the newly allocated buffer
    std::vector<TCHAR> mszExpandedPathList{dwPathListBufferSize, std::allocator<TCHAR>{}};
    Status = PdhExpandWildCardPathH(hDataSource, szWildCardPath, mszExpandedPathList.data(), &dwPathListBufferSize, dwFlags);

    //Parse out the MULTI_SZ strings which form the Counter string array
    if ((Status == ERROR_SUCCESS) && dwPathListBufferSize)
      expandedList = ParseMultiSZString(mszExpandedPathList.data());
  }
  
  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);

  return expandedList;
}

DWORD CPdh::LookupPerfIndexByName(_In_opt_ LPCTSTR szMachineName, _In_ LPCTSTR szName)
{
  DWORD dwIndex{0};
  const PDH_STATUS Status{PdhLookupPerfIndexByName(szMachineName, szName, &dwIndex)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);

  return dwIndex;
}

CPDHString CPdh::LookupPerfNameByIndex(_In_opt_ LPCTSTR szMachineName, _In_ DWORD dwIndex)
{
  //What will be the return value from this function
  CPDHString sName;

  //Note PdhLookupPerfNameByIndex does not seem like being called with szNameBuffer set to nullptr on Windows XP 
  //unlike most of the other Pdh functions. For further details please see the MSDN documentation for PdhLookupPerfNameByIndex

  //First call the function to determine the size of buffer required
  DWORD dwBufferSize{PDH_MAX_COUNTER_NAME};
  std::vector<TCHAR> szBuffer{dwBufferSize, std::allocator<TCHAR>{}};
  PDH_STATUS Status{PdhLookupPerfNameByIndex(szMachineName, dwIndex, szBuffer.data(), &dwBufferSize)};
#pragma warning(suppress: 26472)
  if ((Status == static_cast<PDH_STATUS>(PDH_MORE_DATA)) || (Status == static_cast<PDH_STATUS>(PDH_INSUFFICIENT_BUFFER)))
  {
    //Call again with the newly allocated buffer
    szBuffer.resize(dwBufferSize);
    Status = PdhLookupPerfNameByIndex(szMachineName, dwIndex, szBuffer.data(), &dwBufferSize);
    if (Status == ERROR_SUCCESS)
    {
      //Assign the data into the string sent into us
      sName = szBuffer.data();
    }
  }
  else if (Status == ERROR_SUCCESS)
    sName = szBuffer.data();

  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);

  return sName;
}

void CPdh::SetDefaultRealTimeDataSource(_In_ DWORD dwDataSourceId)
{
  const PDH_STATUS Status{PdhSetDefaultRealTimeDataSource(dwDataSourceId)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    ThrowPdhException(Status);
}


CPdhDataSource::CPdhDataSource() noexcept : m_hDataSource{nullptr}
{
}

CPdhDataSource::~CPdhDataSource()
{
  Close();
}

void CPdhDataSource::Open(_In_ const CPDHStringArray& LogFileNames)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hDataSource == nullptr);

  //Work out the size of the buffer we will need
  std::vector<TCHAR> buffer;
  size_t nSize{0};
  const auto nStrings = LogFileNames.size();
  if (nStrings)
  {
    for (const auto& logFileName : LogFileNames)
      nSize += (logFileName.length() + 1); //1 extra for each null terminator

    //Need one second null for the double null at the end
    ++nSize;

    //Allocate the memory we want
    buffer.resize(nSize);

    //Now copy the strings into the buffer
    size_t nCurOffset{0};
#pragma warning(suppress: 26429)
    LPTSTR lpszString{buffer.data()};
    for (const auto& logFileName : LogFileNames)
    {
      const auto nCurrentStringLength{logFileName.length()};
#pragma warning(suppress: 26481)
      _tcscpy_s(&lpszString[nCurOffset], nCurrentStringLength+1, logFileName.c_str());
      nCurOffset += (nCurrentStringLength + 1);
    }

    //Don't forget to doubly null terminate
#pragma warning(suppress: 26481)
    lpszString[nCurOffset] = _T('\0');
  }

  //Call the SDK function
  const PDH_STATUS Status{PdhBindInputDataSource(&m_hDataSource, buffer.data())};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    CPdh::ThrowPdhException(Status);
}

void CPdhDataSource::Open(_In_ LPCTSTR szLogFileName, _In_ DWORD dwAccessFlags, _Inout_ LPDWORD lpdwLogType, _In_opt_ PDH_HQUERY hQuery, _In_ DWORD dwMaxSize, _In_opt_ LPCTSTR szUserCaption)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hDataSource == nullptr);

  const PDH_STATUS Status{PdhOpenLog(szLogFileName, dwAccessFlags, lpdwLogType, hQuery, dwMaxSize, szUserCaption, &m_hDataSource)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    CPdh::ThrowPdhException(Status);
}

void CPdhDataSource::Close(_In_ DWORD dwFlags) noexcept
{
  if (m_hDataSource != nullptr)
  {
    PdhCloseLog(m_hDataSource, dwFlags);
    m_hDataSource = nullptr;
  }
}

void CPdhDataSource::Update(_In_ LPCTSTR szUserString)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hDataSource != nullptr); //Did you forget to call Open

  const PDH_STATUS Status{PdhUpdateLog(m_hDataSource, szUserString)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    CPdh::ThrowPdhException(Status);
}

LONGLONG CPdhDataSource::GetSize()
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hDataSource != nullptr); //Did you forget to call Open

  LONGLONG nSize{0};
  const PDH_STATUS Status{PdhGetLogFileSize(m_hDataSource, &nSize)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    CPdh::ThrowPdhException(Status);

  return nSize;
}

PPDH_RAW_LOG_RECORD CPdhDataSource::ReadRawRecord(_In_ FILETIME ftRecord)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hDataSource != nullptr); //Did you forget to call Open

  //What will be the return value from this function
  PPDH_RAW_LOG_RECORD pRawLogRecord{nullptr};

  //First call the function to determine the size of buffer required
  DWORD dwBufferSize{0};
  PDH_STATUS Status{PdhReadRawLogRecord(m_hDataSource, ftRecord, nullptr, &dwBufferSize)};
#pragma warning(suppress: 26472)
  if (Status == static_cast<PDH_STATUS>(PDH_MORE_DATA))
  {
    //Call again with the newly allocated buffer (client code is expected to do a DeallocateMemory on the returned value)
    pRawLogRecord = static_cast<PPDH_RAW_LOG_RECORD>(AllocateMemory(dwBufferSize * sizeof(TCHAR)));
    if (pRawLogRecord == nullptr)
      CPdh::ThrowPdhException(PDH_MEMORY_ALLOCATION_FAILURE);
    Status = PdhReadRawLogRecord(m_hDataSource, ftRecord, pRawLogRecord, &dwBufferSize);
  }
  
  if (Status != ERROR_SUCCESS)
  {
    //Free up the memory before we throw the exception
    if (pRawLogRecord != nullptr)
      DeallocateMemory(pRawLogRecord);
    CPdh::ThrowPdhException(Status);
  }

#pragma warning(suppress: 6001)
  return pRawLogRecord;
}

LPVOID CPdhDataSource::AllocateMemory(_In_ DWORD dwBytes) noexcept
{
  //By default we allocate from the Process Heap
  return HeapAlloc(GetProcessHeap(), 0, dwBytes);
}

void CPdhDataSource::DeallocateMemory(__drv_freesMem(Mem) _Frees_ptr_opt_ void* pData) noexcept
{
  //By default we deallocate from the Process Heap
  HeapFree(GetProcessHeap(), 0, pData);
}

CPDHString CPdhDataSource::GetDefaultPerfCounter(_In_opt_ LPCTSTR szMachineName, _In_ LPCTSTR  szObjectName)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hDataSource != nullptr); //Did you forget to call Open

  //What will be the return value from this method
  CPDHString sDefaultCounterName;

  //First call the function to determine the size of buffer required
  DWORD dwBufferSize{0};
  PDH_STATUS Status{PdhGetDefaultPerfCounterH(m_hDataSource, szMachineName, szObjectName, nullptr, &dwBufferSize)};
#pragma warning(suppress: 26472)
  if (Status == static_cast<PDH_STATUS>(PDH_MORE_DATA))
  {
    //Call again with the newly allocated buffer
    std::vector<TCHAR> pszBuffer{dwBufferSize, std::allocator<TCHAR>{}};
    Status = PdhGetDefaultPerfCounterH(m_hDataSource, szMachineName, szObjectName, pszBuffer.data(), &dwBufferSize);
    if (Status == ERROR_SUCCESS)
    {
      //Assign the data into the string sent into us
      sDefaultCounterName = pszBuffer.data();
    }
  }

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    CPdh::ThrowPdhException(Status);

  return sDefaultCounterName;
}

void CPdhDataSource::UpdateLogFileCatalog()
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hDataSource != nullptr); //Did you forget to call Open

  const PDH_STATUS Status{PdhUpdateLogFileCatalog(m_hDataSource)};

  //Throw an exception if anything goes wrong
  if (Status != ERROR_SUCCESS)
    CPdh::ThrowPdhException(Status);
}
